@@ -4,7 +4,7 @@ use crate::PlatformTag;
 use crate::auditwheel::AuditWheelMode;
 use anyhow::{Context, Result};
 use fs_err as fs;
-use pep440_rs::Version;
+use pep440_rs::{Version, VersionSpecifiers};
 use pep508_rs::VersionOrUrl;
 use pyproject_toml::{BuildSystem, Project};
 use serde::{Deserialize, Serialize};
@@ -231,6 +231,124 @@ pub struct SbomConfig {
     pub include: Option<Vec<PathBuf>>,
 }
 
+/// A cargo feature specification that can be either a plain feature name
+/// or a conditional feature that is only enabled for certain Python versions.
+///
+/// # Examples
+///
+/// ```toml
+/// [tool.maturin]
+/// features = [
+///   "some-feature",
+///   { feature = "pyo3/abi3-py311", python-version = ">=3.11" },
+///   { feature = "pyo3/abi3-py38", python-version = "<3.11" },
+///   { feature = "pyo3/abi3-py311", python-version = ">=3.11", python-implementation = "cpython" },
+///   { feature = "pypy-compat", python-implementation = "pypy" },
+/// ]
+/// ```
+#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq)]
+#[serde(untagged)]
+#[cfg_attr(feature = "schemars", derive(schemars::JsonSchema))]
+pub enum FeatureSpec {
+    /// A plain feature name, always enabled
+    Plain(String),
+    /// A feature enabled only when the conditions match
+    Conditional {
+        /// The cargo feature to enable
+        feature: String,
+        /// PEP 440 version specifier for the target Python version, e.g. ">=3.11"
+        #[serde(
+            rename = "python-version",
+            default,
+            skip_serializing_if = "Option::is_none"
+        )]
+        #[cfg_attr(feature = "schemars", schemars(with = "Option<String>"))]
+        python_version: Option<VersionSpecifiers>,
+        /// Python implementation name, e.g. "cpython", "pypy", "graalpy"
+        #[serde(
+            rename = "python-implementation",
+            default,
+            skip_serializing_if = "Option::is_none"
+        )]
+        python_implementation: Option<String>,
+    },
+}
+
+/// A conditional feature with its matching criteria.
+#[derive(Debug, Clone, PartialEq, Eq)]
+pub struct ConditionalFeature {
+    /// The cargo feature to enable
+    pub feature: String,
+    /// PEP 440 version specifier for the target Python version
+    pub python_version: Option<VersionSpecifiers>,
+    /// Python implementation name, e.g. "cpython", "pypy", "graalpy"
+    pub python_implementation: Option<String>,
+}
+
+impl FeatureSpec {
+    /// Split a list of feature specs into plain features and conditional features.
+    pub fn split(specs: Vec<FeatureSpec>) -> (Vec<String>, Vec<ConditionalFeature>) {
+        let mut plain = Vec::new();
+        let mut conditional = Vec::new();
+        for spec in specs {
+            match spec {
+                FeatureSpec::Plain(f) => plain.push(f),
+                FeatureSpec::Conditional {
+                    feature,
+                    python_version: None,
+                    python_implementation: None,
+                } => plain.push(feature),
+                FeatureSpec::Conditional {
+                    feature,
+                    python_version,
+                    python_implementation,
+                } => conditional.push(ConditionalFeature {
+                    feature,
+                    python_version,
+                    python_implementation,
+                }),
+            }
+        }
+        (plain, conditional)
+    }
+
+    /// Resolve which conditional features should be enabled for the given
+    /// build environment.
+    pub fn resolve_conditional(
+        conditional_features: &[ConditionalFeature],
+        env: &FeatureConditionEnv,
+    ) -> Vec<String> {
+        let python_version = Version::new([env.major as u64, env.minor as u64]);
+        conditional_features
+            .iter()
+            .filter(|f| {
+                f.python_version
+                    .as_ref()
+                    .is_none_or(|s| s.contains(&python_version))
+            })
+            .filter(|f| {
+                f.python_implementation
+                    .as_ref()
+                    .is_none_or(|i| i.eq_ignore_ascii_case(env.implementation_name))
+            })
+            .map(|f| f.feature.clone())
+            .collect()
+    }
+}
+
+/// The build environment used to evaluate conditional features.
+///
+/// TODO: add fields like `target_os`, `target_arch`, etc. for
+/// Rust target-based conditions.
+pub struct FeatureConditionEnv<'a> {
+    /// Python major version
+    pub major: usize,
+    /// Python minor version
+    pub minor: usize,
+    /// Python implementation name, e.g. "cpython", "pypy"
+    pub implementation_name: &'a str,
+}
+
 /// The `[tool.maturin]` section of a pyproject.toml
 #[derive(Serialize, Deserialize, Debug, Clone, Default)]
 #[serde(rename_all = "kebab-case")]
@@ -279,8 +397,10 @@ pub struct ToolMaturin {
     pub profile: Option<String>,
     /// Same as `profile` but for "editable" builds
     pub editable_profile: Option<String>,
-    /// Space or comma separated list of features to activate
-    pub features: Option<Vec<String>>,
+    /// List of features to activate.
+    /// Each entry can be a plain feature name string, or a conditional object
+    /// with `feature` and `python-version` keys.
+    pub features: Option<Vec<FeatureSpec>>,
     /// Activate all available features
     pub all_features: Option<bool>,
     /// Do not activate the `default` feature
@@ -552,7 +672,9 @@ impl PyProjectToml {
 mod tests {
     use crate::{
         PyProjectToml,
-        pyproject_toml::{Format, Formats, GlobPattern, ToolMaturin},
+        pyproject_toml::{
+            FeatureConditionEnv, FeatureSpec, Format, Formats, GlobPattern, ToolMaturin,
+        },
     };
     use expect_test::expect;
     use fs_err as fs;
@@ -599,7 +721,10 @@ mod tests {
         assert_eq!(maturin.profile.as_deref(), Some("dev"));
         assert_eq!(
             maturin.features,
-            Some(vec!["foo".to_string(), "bar".to_string()])
+            Some(vec![
+                FeatureSpec::Plain("foo".to_string()),
+                FeatureSpec::Plain("bar".to_string()),
+            ])
         );
         assert!(maturin.all_features.is_none());
         assert_eq!(maturin.no_default_features, Some(true));
@@ -813,4 +938,151 @@ mod tests {
         "#]];
         expected.assert_eq(&inner_error.to_string());
     }
+
+    #[test]
+    fn test_resolve_conditional_features() {
+        let specs = vec![
+            FeatureSpec::Conditional {
+                feature: "pyo3/abi3-py311".to_string(),
+                python_version: Some(">=3.11".parse().unwrap()),
+                python_implementation: None,
+            },
+            FeatureSpec::Conditional {
+                feature: "pyo3/abi3-py38".to_string(),
+                python_version: Some("<3.11".parse().unwrap()),
+                python_implementation: None,
+            },
+            FeatureSpec::Conditional {
+                feature: "fast-buffer".to_string(),
+                python_version: Some(">=3.11".parse().unwrap()),
+                python_implementation: None,
+            },
+        ];
+        let (_plain, conditional) = FeatureSpec::split(specs);
+
+        let cpython = |major, minor| FeatureConditionEnv {
+            major,
+            minor,
+            implementation_name: "cpython",
+        };
+
+        // Python 3.12 should match >=3.11
+        let resolved = FeatureSpec::resolve_conditional(&conditional, &cpython(3, 12));
+        assert_eq!(resolved, vec!["pyo3/abi3-py311", "fast-buffer"]);
+
+        // Python 3.11 should match >=3.11
+        let resolved = FeatureSpec::resolve_conditional(&conditional, &cpython(3, 11));
+        assert_eq!(resolved, vec!["pyo3/abi3-py311", "fast-buffer"]);
+
+        // Python 3.9 should match <3.11
+        let resolved = FeatureSpec::resolve_conditional(&conditional, &cpython(3, 9));
+        assert_eq!(resolved, vec!["pyo3/abi3-py38"]);
+
+        // Python 3.8 should match <3.11
+        let resolved = FeatureSpec::resolve_conditional(&conditional, &cpython(3, 8));
+        assert_eq!(resolved, vec!["pyo3/abi3-py38"]);
+    }
+
+    #[test]
+    fn test_resolve_conditional_features_with_implementation() {
+        let specs = vec![
+            FeatureSpec::Conditional {
+                feature: "pyo3/abi3-py311".to_string(),
+                python_version: Some(">=3.11".parse().unwrap()),
+                python_implementation: Some("cpython".to_string()),
+            },
+            FeatureSpec::Conditional {
+                feature: "pypy-compat".to_string(),
+                python_version: None,
+                python_implementation: Some("pypy".to_string()),
+            },
+            FeatureSpec::Conditional {
+                feature: "always-conditional".to_string(),
+                python_version: Some(">=3.10".parse().unwrap()),
+                python_implementation: None,
+            },
+        ];
+        let (_plain, conditional) = FeatureSpec::split(specs);
+
+        let env = |major, minor, impl_name| FeatureConditionEnv {
+            major,
+            minor,
+            implementation_name: impl_name,
+        };
+
+        // CPython 3.12 should match abi3 and always-conditional
+        let resolved = FeatureSpec::resolve_conditional(&conditional, &env(3, 12, "cpython"));
+        assert_eq!(resolved, vec!["pyo3/abi3-py311", "always-conditional"]);
+
+        // PyPy 3.10 should match pypy-compat and always-conditional
+        let resolved = FeatureSpec::resolve_conditional(&conditional, &env(3, 10, "pypy"));
+        assert_eq!(resolved, vec!["pypy-compat", "always-conditional"]);
+
+        // CPython 3.10 should only match always-conditional (not abi3 due to version)
+        let resolved = FeatureSpec::resolve_conditional(&conditional, &env(3, 10, "cpython"));
+        assert_eq!(resolved, vec!["always-conditional"]);
+
+        // PyPy 3.9 should only match pypy-compat (version too low for always-conditional)
+        let resolved = FeatureSpec::resolve_conditional(&conditional, &env(3, 9, "pypy"));
+        assert_eq!(resolved, vec!["pypy-compat"]);
+    }
+
+    #[test]
+    fn test_feature_spec_deserialize_mixed() {
+        let toml_str = r#"
+            features = [
+                "plain-feature",
+                { feature = "pyo3/abi3-py311", python-version = ">=3.11" },
+            ]
+        "#;
+        let maturin: ToolMaturin = toml::from_str(toml_str).unwrap();
+        assert_eq!(
+            maturin.features,
+            Some(vec![
+                FeatureSpec::Plain("plain-feature".to_string()),
+                FeatureSpec::Conditional {
+                    feature: "pyo3/abi3-py311".to_string(),
+                    python_version: Some(">=3.11".parse().unwrap()),
+                    python_implementation: None,
+                },
+            ])
+        );
+    }
+
+    #[test]
+    fn test_feature_spec_deserialize_with_implementation() {
+        let toml_str = r#"
+            features = [
+                { feature = "pyo3/abi3-py311", python-version = ">=3.11", python-implementation = "cpython" },
+                { feature = "pypy-compat", python-implementation = "pypy" },
+            ]
+        "#;
+        let maturin: ToolMaturin = toml::from_str(toml_str).unwrap();
+        assert_eq!(
+            maturin.features,
+            Some(vec![
+                FeatureSpec::Conditional {
+                    feature: "pyo3/abi3-py311".to_string(),
+                    python_version: Some(">=3.11".parse().unwrap()),
+                    python_implementation: Some("cpython".to_string()),
+                },
+                FeatureSpec::Conditional {
+                    feature: "pypy-compat".to_string(),
+                    python_version: None,
+                    python_implementation: Some("pypy".to_string()),
+                },
+            ])
+        );
+    }
+
+    #[test]
+    fn test_feature_spec_deserialize_invalid_specifier() {
+        let toml_str = r#"
+            features = [
+                { feature = "foo", python-version = "not-a-version" },
+            ]
+        "#;
+        let result: Result<ToolMaturin, _> = toml::from_str(toml_str);
+        assert!(result.is_err());
+    }
 }
