@@ -1223,7 +1223,7 @@ fn pyo3_source_date_epoch() {
     unsafe { env::set_var("SOURCE_DATE_EPOCH", "0") };
     handle_result(other::check_wheel_mtimes(
         "test-crates/pyo3-mixed-include-exclude",
-        vec![datetime!(1980-01-01 0:00 UTC)],
+        vec![datetime!(1980-01-01 0:00)],
         "pyo3_source_date_epoch",
     ))
 }
